import logging

from typing import List

import datetime
from sqlalchemy import Column, Integer, Text, DATETIME, ForeignKey, Float
from sqlalchemy.orm import relationship, Mapped

import sys


from JuMonC import settings
from JuMonC.models.cache.database import Base, db_session
from JuMonC._version import __DB_version__


logger = logging.getLogger(__name__)

class CacheEntry(Base):
    __tablename__ = 'cache_entry'
    query = db_session.query_property()
    cache_id = Column('cache_id', Integer, primary_key=True)
    time = Column('time', DATETIME, unique=False)
    API_path = Column('API_path', Text, unique=False)
    
    parameters:Mapped[List["Parameter"]] = relationship("Parameter",
                                  back_populates="cache_entry",
                                  cascade="all, delete",
                                  passive_deletes=True,
                                  uselist=True
                             )
    
    results_int:Mapped[List["ResultInt"]] = relationship("ResultInt",
                                  back_populates="cache_entry",
                                  cascade="all, delete",
                                  passive_deletes=True,
                                  uselist=True
                             )
    
    results_float:Mapped[List["ResultFloat"]] = relationship("ResultFloat",
                                  back_populates="cache_entry",
                                  cascade="all, delete",
                                  passive_deletes=True,
                                  uselist=True
                             )
    
    results_str:Mapped[List["ResultStr"]] = relationship("ResultStr",
                                  back_populates="cache_entry",
                                  cascade="all, delete",
                                  passive_deletes=True,
                                  uselist=True
                             )

    def __init__(self, API_path:str) -> None:
        """DB class!"""
        self.time = datetime.datetime.now()
        self.API_path = API_path

    def __repr__(self) -> str:
        """DB to string method."""
        if self.time:
            return f'<cache_id {self.cache_id!r}, time {self.time.strftime(settings.DATETIME_FORMAT)!r}, API_path {self.API_path!r}>'
        logging.warning("There is a time value missing in the database for ID: %i", self.cache_id)
        return ""

    
class Parameter(Base):
    __tablename__ = 'parameter'
    query = db_session.query_property()
    cache_id = Column('cache_id', Integer, ForeignKey('cache_entry.cache_id', ondelete="CASCADE"), primary_key=True)
    parameter_name = Column('parameter_name', Text, unique=False, primary_key=True)
    parameter_value = Column('parameter_value', Text, unique=False)
    
    cache_entry:Mapped[CacheEntry] = relationship("CacheEntry", back_populates="parameters", uselist=False)

    def __init__(self, cache_id:int, parameter_name:str, parameter_value:str) -> None:
        """DB class!"""
        self.cache_id = cache_id
        self.parameter_name = parameter_name
        self.parameter_value = parameter_value

    def __repr__(self) -> str:
        """DB to string method."""
        return (f'<cache_id {self.cache_id!r}, '
                f'parameter_name {self.parameter_name!r}, parameter_value {self.parameter_value!r}>')

    
class ResultInt(Base):
    __tablename__ = 'results_int'
    query = db_session.query_property()
    cache_id = Column('cache_id', Integer, ForeignKey('cache_entry.cache_id', ondelete="CASCADE"), primary_key=True)
    result_name = Column('result_name', Text, unique=False, primary_key=True)
    result = Column('result', Integer, unique=False)
    
    cache_entry:Mapped[CacheEntry] = relationship("CacheEntry", back_populates="results_int", uselist=False)

    def __init__(self, cache_id:int, result_name:str, result:int) -> None:
        """DB class!"""
        self.cache_id = cache_id
        self.result_name = result_name
        self.result = result

    def __repr__(self) -> str:
        """DB to string method."""
        return f'<cache_id {self.cache_id!r}, result_name {self.result_name!r}, result {self.result!r}>'

    
class ResultFloat(Base):
    __tablename__ = 'results_float'
    query = db_session.query_property()
    cache_id = Column('cache_id', Integer, ForeignKey('cache_entry.cache_id', ondelete="CASCADE"), primary_key=True)
    result_name = Column('result_name', Text, unique=False, primary_key=True)
    result = Column('result', Float, unique=False)
    
    cache_entry:Mapped[CacheEntry] = relationship("CacheEntry", back_populates="results_float", uselist=False)

    def __init__(self, cache_id:int, result_name:str, result:float) -> None:
        """DB class!"""
        self.cache_id = cache_id
        self.result_name = result_name
        self.result = result

    def __repr__(self) -> str:
        """DB to string method."""
        return f'<cache_id {self.cache_id!r}, result_name {self.result_name!r}, result {self.result!r}>'

    
class ResultStr(Base):
    __tablename__ = 'results_str'
    query = db_session.query_property()
    cache_id = Column('cache_id', Integer, ForeignKey('cache_entry.cache_id', ondelete="CASCADE"), primary_key=True)
    result_name = Column('result_name', Text, unique=False, primary_key=True)
    result = Column('result', Text, unique=False)
    
    cache_entry:Mapped[CacheEntry] = relationship("CacheEntry", back_populates="results_str", uselist=False)

    def __init__(self, cache_id:int, result_name:str, result:str) -> None:
        """DB class!"""
        self.cache_id = cache_id
        self.result_name = result_name
        self.result = result

    def __repr__(self) -> str:
        """DB to string method."""
        return f'<cache_id {self.cache_id!r}, result_name {self.result_name!r}, result {self.result!r}>'

    
class Version(Base):
    __tablename__ = 'version'
    query = db_session.query_property()
    entry = Column('entry', Text, unique=True, primary_key=True)
    value = Column('value', Text, unique=False)

    def __init__(self, entry:str, value:str) -> None:
        """DB class!"""
        self.entry = entry
        self.value = value

    def __repr__(self) -> str:
        """DB to string method."""
        return f'<entry {self.entry!r}, value {self.value!r}>'
    
def check_db_version() -> None:
    logging.debug("Checking DB version")
    db_name = Version.query.filter_by(entry="name").first()
    if db_name:
        if db_name.value == "JuMonC":
            logging.debug("Version table contains the name JuMonC")
        else:
            logging.error(("Version table in db contains the name: %s, that is unexpected."
                            " To prevent JuMonC deleting the database of another projekt,"
                            " JuMonC will stop. DB_PATH: %s"), db_name.name, settings.DB_PATH)
            sys.exit(-2)
    else:
        db_name = Version("name", "JuMonC")
        db_session.add(db_name)
    
    db_version = Version.query.filter_by(entry="DB_version").first()
    if db_version:
        if db_version.value == __DB_version__:
            logging.info("Using exsiting DB with db version: %s", __DB_version__)
        else:
            logging.error(("Version table in db contains the version: %s, that is unexpected."
                            " To prevent JuMonC causing errors in an older version,"
                            " JuMonC will stop. DB_PATH: %s"), db_version.value, settings.DB_PATH)
            sys.exit(-2)
    else:
        db_version = Version("DB_version", __DB_version__)
        logging.info("Creating DB with db version: %s", __DB_version__)
        db_session.add(db_version)
    
    db_session.commit()


import logging

from typing import Union, Tuple


logger = logging.getLogger(__name__)



_conversion_binary_dict = {
2**50: "Pi",
2**40: "Ti",
2**30: "Gi",
2**20: "Mi",
2**10: "Ki",
}



def convertBinaryPrefix(value: Union[int, float], decimalDigits: int = 2) -> Tuple[str, str]:
    for (key, unitPrefix) in _conversion_binary_dict.items():
        if value >= 10*key:
            value = value/key
            return (("{:." + str(decimalDigits) + "f}").format(value), unitPrefix)
    if isinstance(value, int):
        return (str(value), "")
    return (("{:." + str(decimalDigits) + "f}").format(value), "")


import logging

from flask import abort, jsonify, make_response, Response, request

from typing import Dict, List, Union


from JuMonC.handlers.base import api_version_path, check_version, RESTAPI
from JuMonC.authentication import scopes
from JuMonC.authentication.check import check_auth
from JuMonC.models import pluginInformation
from JuMonC.tasks import disk
from JuMonC import settings

import JuMonC.models.cache.helper as cache


logger = logging.getLogger(__name__)

links:        List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []
status_links: List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []
config_links: List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []

io_path = "/io"

@RESTAPI.route(api_version_path + io_path, methods=["GET"])
@check_version
@check_auth(scopes["see_links"])
def returnIOLinks(version: int) -> Response:
    logging.debug("Accessed /v%i/io/", version)
    return make_response(jsonify(sorted(links, key=lambda dic: dic['link'])), 200)

@RESTAPI.route(api_version_path + io_path + "/status", methods=["GET"])
@check_version
@check_auth(scopes["see_links"])
def returnIOStatusLinks(version: int) -> Response:
    logging.debug("Accessed /v%i/io/status", version)
    return make_response(jsonify(sorted(status_links, key=lambda dic: dic['link'])), 200)

@RESTAPI.route(api_version_path + io_path + "/config", methods=["GET"])
@check_version
@check_auth(scopes["see_links"])
def returnIOConfigLinks(version: int) -> Response:
    logging.debug("Accessed /v%i/io/config", version)
    return make_response(jsonify(sorted(config_links, key=lambda dic: dic['link'])), 200)

def registerRestApiPaths(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    links.append(registerStatusLinks(version))
    #links.append(registerConfigLinks(version)) not in use for now
    return {
        "link": "/v" + str(version) + io_path,
        "isOptional": False,
        "description": "Gather information about the disk IO",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }


def registerStatusLinks(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    types : List[str] = ["read_count", "write_count",
                         "read_bytes", "write_bytes",
                         "read_time", "write_time",
                         "busy_time",
                         "read_merged_count", "write_merged_count"]
    descriptions : List[str] = ["number of reads", "number of writes",
                                "number of bytes read", "number of bytes written",
                                "time spent reading from disk", "time spent writing to disk" ,
                                "time spent doing actual I/Os",
                                "number of merged reads (see iostats doc)", "number of merged writes (see iostats doc)"]
    if pluginInformation.disk_is_working:
        for i, typeStr in enumerate(types):
            status_links.append({
            "link": "/v" + str(version) + io_path + "/status/" + typeStr,
            "isOptional": True,
            "description": descriptions[i],
            "parameters": [
                {"name": "duration",
                "description": "When this parameter is not present the raw data value(s) will be provided, otherwise the average value per second"},
                {"name": "humanReadable",
                "description": "For 'True' convert to better readable numbers, for 'False' return actual number. " + 
                 "If not set to valid value, uses default value(" + str(settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS) + ")."},
                {"name": "node",
                "description": "When this parameter is not present the average data of all nodes is presented, otherwise only from the choosen node"},
                {"name": "token",
                "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
            })
        
        @RESTAPI.route(api_version_path + io_path + "/status/<string:datatype>", methods=["GET"])
        @check_version
        @check_auth(scopes["compute_data"])
        def returnIOStatus(version: int, datatype:str) -> Response:
            duration = request.args.get('duration', default = -1.0, type = float)
            humanReadable = request.args.get('humanReadable', default = settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS, type = settings.helpers.parse_boolean)
            logging.debug("Accessed /v%i/io/status/%s", version, datatype)
            
            cache_id = cache.add_cache_entry("/v" + str(version) + "/io/status/" + datatype)
            cache.addParameter(cache_id, "duration", str(duration))
            
            data = disk.plugin.getStatusData(dataType = datatype,
                                               duration = duration,
                                               overrideHumanReadableWithValue = humanReadable)
            if len(data) > 0:
                for data_element in data:
                    for result_name, result in data_element.items():
                        cache.addResult(cache_id, result_name, result)
                return make_response(jsonify(data), 200)
            
            logging.warning("Accessed /v%i/io/status/%s, but not avaiable", version, datatype)
            abort(404)
            return make_response("",404)
        
    return {
        "link": "/v" + str(version) + io_path + "/status",
        "isOptional": False,
        "description": "Gather information concerning the memory status",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }


def registerConfigLinks(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    types : List[str] = []
    descriptions : List[str] = []
    if pluginInformation.memory_is_working:
        for i, typeStr in enumerate(types):
            config_links.append({
            "link": "/v" + str(version) + io_path + "/status/" + typeStr,
            "isOptional": True,
            "description": descriptions[i],
            "parameters": [
                {"name": "humanReadable",
                "description": "For 'True' convert to better readable numbers, for 'False' return actual number. " + 
                 "If not set to valid value, uses default value(" + str(settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS) + ")."},
                {"name": "node",
                "description": "When this parameter is not present the average data of all nodes is presented, otherwise only from the choosen node"},
                {"name": "token",
                "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
            })
        
        @RESTAPI.route(api_version_path + io_path + "/config/<string:datatype>", methods=["GET"])
        @check_version
        @check_auth(scopes["compute_data"])
        def returnIOConfig(version: int, datatype:str) -> Response:
            humanReadable = request.args.get('humanReadable', default = settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS, type = settings.helpers.parse_boolean)
            logging.debug("Accessed /v%i/io/config/%s", version, datatype)
            data = disk.plugin.getConfigData(dataType = datatype, overrideHumanReadableWithValue = humanReadable)
            if len(data) > 0:
                return make_response(jsonify(data), 200)
            
            logging.warning("Accessed /v%i/io/config/%s, but not avaiable", version, datatype)
            abort(404)
            return make_response("",404)
        
    return {
        "link": "/v" + str(version) + io_path + "/config",
        "isOptional": False,
        "description": "Gather information concerning the memory config",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }



import logging

from typing import Dict, List, Union, Optional
from types import ModuleType

from JuMonC.tasks import Plugin
from JuMonC.helpers import convertNumbers
from JuMonC import settings


logger = logging.getLogger(__name__)


class _memoryPlugin(Plugin.Plugin): 
    
    _psutil: Optional[ModuleType]
    
    def __init__(self) -> None:
        super().__init__()
        self.isWorking()
    
    
    def _isWorking(self) -> bool:
        try:
            self._psutil = __import__('psutil')
        except ModuleNotFoundError:
            logging.warning("psutil can not be imported, therefore memory functionality not avaiable")
            return False
        return True
    
    def getFree(self) -> float:
        if isinstance(self._psutil, ModuleType):
            return self._psutil.virtual_memory().free
        logging.error("Something went wrong, in memory plugin psutil is not ModuleType")
        return -1.0
    
    def getUsed(self) -> float:
        if isinstance(self._psutil, ModuleType):
            return self._psutil.virtual_memory().used
        logging.error("Something went wrong, in memory plugin psutil is not ModuleType")
        return -1.0
    
    def getTotal(self) -> float:
        if isinstance(self._psutil, ModuleType):
            return self._psutil.virtual_memory().total
        logging.error("Something went wrong, in memory plugin psutil is not ModuleType")
        return -1.0
    
    
    def getStatusData(self,
                      dataType: str, 
                      duration: float = -1.0, 
                      overrideHumanReadableWithValue: Optional[bool] = None) -> List[Dict[str, Union[bool, int, float ,str]]]:
        logging.debug("get memory status data with type=%s, duration=%s, overrride humand readable=%s",
                     str(dataType),
                     str(duration),
                     str(overrideHumanReadableWithValue))
        if dataType == "free":
            return [self.convertNumber(self.getFree(), "free", "B", overrideHumanReadableWithValue)]
        if dataType == "used":
            return [self.convertNumber(self.getUsed(), "used", "B", overrideHumanReadableWithValue)]
        return []
    
    
    def getConfigData(self,
                      dataType: str, 
                      overrideHumanReadableWithValue: Optional[bool] = None) -> List[Dict[str, Union[bool, int, float ,str]]]:
        logging.debug("get memory config data with type=%s, overrride humand readable=%s",
                     str(dataType),
                     str(overrideHumanReadableWithValue))
        if dataType == "total":
            return [self.convertNumber(self.getTotal(), "total", "B", overrideHumanReadableWithValue)]
        return []
    
    
    def convertNumber(self, 
                      num: Union[int,float], 
                      name: str, 
                      unit: str, 
                      overrideHumanReadableWithValue: Optional[bool]
                ) -> Dict[str, Union[bool, int, float ,str]]:
        if overrideHumanReadableWithValue or (overrideHumanReadableWithValue is None and settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS) :
            (value, unitPrefix) = convertNumbers.convertBinaryPrefix(num)
            return {name + "[" + unitPrefix + unit + "]" : value}
        
        return {name + "[" + unit + "]" : str(num)}
    
    
plugin = _memoryPlugin()


import logging
logger = logging.getLogger(__name__)


def parse_boolean(s: str) -> bool:
    """Takes a string and returns the equivalent as a boolean value."""
    s = s.strip().lower()
    if s in ("yes", "true", "on", "1"):
        return True
    if s in ("no", "false", "off", "0", "none"):
        return False
    raise ValueError(f'Invalid boolean value {s}')



slurm_is_working = False

papi_is_working = False

nvidia_is_working = False

network_is_working = False

disk_is_working = False

cpu_is_working = False

memory_is_working = False


flask_path_reg = False


import logging

from typing import Dict, List, Union, Optional
from types import ModuleType

from JuMonC.tasks import Plugin


logger = logging.getLogger(__name__)


class _CPUPlugin(Plugin.Plugin): 
    
    _psutil: Optional[ModuleType]
    _load_length = {1 : 0, 5: 1, 15: 2}
    
    def __init__(self) -> None:
        super().__init__()
        if self.isWorking():
            self.getLoad(1) # make sure it is correctly init even for windows
    
    
    def _isWorking(self) -> bool:
        try:
            self._psutil = __import__('psutil')
        except ModuleNotFoundError:
            logging.warning("psutil can not be imported, therefore CPU functionality not avaiable")
            return False
        return True
    
    def getLoad(self, length: int = 1) -> float:
        if isinstance(self._psutil, ModuleType):
            return self._psutil.getloadavg()[self._load_length[length]]
        logging.error("Something went wrong, in cpu plugin psutil is not ModuleType")
        return -1.0
    
    
    def getStatusData(self,
                      dataType: str, 
                      duration: float, 
                      overrideHumanReadableWithValue: Optional[bool] = None) -> List[Dict[str, Union[bool, int, float ,str]]]:
        logging.debug("get CPU status data with type=%s, duration=%s, overrride humand readable=%s",
                     str(dataType),
                     str(duration),
                     str(overrideHumanReadableWithValue))
        if dataType == "load":
            return [{"load": self.getLoad(int(duration))}]
        return []
    
    
    def getConfigData(self,
                      dataType: str, 
                      overrideHumanReadableWithValue: Optional[bool] = None) -> List[Dict[str, Union[bool, int, float ,str]]]:
        logging.debug("get CPU config data with type=%s, overrride humand readable=%s",
                     str(dataType),
                     str(overrideHumanReadableWithValue))
        return []
    
    
plugin = _CPUPlugin()


import logging

from flask import jsonify, make_response, Response

from typing import Dict, List, Union


from JuMonC.handlers import base
from JuMonC.handlers import versionTree
from JuMonC.handlers import auth


logger = logging.getLogger(__name__)

links: List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []




@base.RESTAPI.route("/", methods=["GET"])
def returnRootLinks() -> Response:
    return make_response(jsonify(sorted(links, key=lambda dic: dic['link'])), 200)


def registerRestApiPaths() -> None:
    for i in range(base.start_version, base.end_version + 1):
        links.append(versionTree.registerRestApiPaths(i))
        
    links.append(auth.registerLogin())
    links.append(auth.registerLogout())
    links.append(auth.registerScope())

    links.append({
        "link": "/version",
        "isOptional": False,
        "description": "Get current version of JuMonC",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    })
    links.append({
        "link": "/ping",
        "isOptional": False,
        "description": "Test link to test avaibility of JuMonC REST API",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    })
    links.append({
        "link": "/stop",
        "isOptional": False,
        "description": "Calling this link stops JuMonC",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    })


import logging

from typing import Union, Optional

from JuMonC.models.cache.database import db_session
import JuMonC.models.cache.dbmodel as cache_model


logger = logging.getLogger(__name__)


def add_cache_entry(API_path:str) -> Optional[int]:
    entry = cache_model.CacheEntry(API_path)
    db_session.add(entry)
    db_session.commit()
    
    return entry.cache_id


def addParameter(cache_id:Optional[int], parameter_name:str, parameter_value:str) -> None:
    if not cache_id:
        logging.error("Missing cache_id")
        return
    
    parameter = cache_model.Parameter(cache_id, parameter_name, parameter_value)
    db_session.add(parameter)
    db_session.commit()

    
def addResult(cache_id:Optional[int], result_name:str, result:Union[int,float,str]) -> None:
    if not cache_id:
        logging.error("Missing cache_id")
        return
    
    result_entry:Union[cache_model.ResultInt, cache_model.ResultFloat, cache_model.ResultStr]
    if isinstance(result, int):
        result_entry = cache_model.ResultInt(cache_id, result_name, result)
        logging.debug("Adding int result to cache databsae")
    elif isinstance(result, float):
        result_entry = cache_model.ResultFloat(cache_id, result_name, result)
        logging.debug("Adding float result to cache databsae")
    elif isinstance(result, str):
        result_entry = cache_model.ResultStr(cache_id, result_name, result)
        logging.debug("Adding str result to cache databsae")
    else:
        logging.warning("Non valid result is not added to cache databsae: %s", str(type(result)))
        return
    db_session.add(result_entry)
    db_session.commit()


import logging

from concurrent.futures import ThreadPoolExecutor

from functools import wraps

from typing import Any, Callable, Optional


from JuMonC import settings


logger = logging.getLogger(__name__)

_thread_pool: Optional[ThreadPoolExecutor] = None
_task_pool_avaiable = False



def setupTaskPool() -> None:
    global _thread_pool
    global _task_pool_avaiable
    if _task_pool_avaiable:
        return
    logging.info("started task pool with a maximum of %s workers", str(settings.MAX_WORKER_THREADS))
    _thread_pool = ThreadPoolExecutor(settings.MAX_WORKER_THREADS)
    _task_pool_avaiable = True

    
def executeAsTask(func: Callable[..., None] ) -> Callable[..., None]:
    @wraps(func)
    def decorated_function(*args: Any, **kwargs: Any) -> None:
        addTask(func, *args, **kwargs)
        
    return decorated_function
        
def addTask(func: Callable, *args: Any, **kwargs: Any) -> None:
    setupTaskPool()
    if _thread_pool is None:
        logging.error("Thread Pool should be avaiable")
        return
    _thread_pool.submit(func, *args, **kwargs)


"""settings package of JuMonC, central place for settings."""


import os

from typing import List, Optional, Tuple, Union

from JuMonC.settings.helpers import parse_boolean

LOG_LEVEL: str = os.environ.get("JUMONC_LOG_LEVEL", "INFO")
LOG_STDOUT: bool = parse_boolean(os.environ.get("JUMONC_LOG_STDOUT", "false"))
LOG_PREFIX: str = os.environ.get("JUMONC_LOG_PREFIX", "")
LOG_FORMAT: str = os.environ.get(
    "JUMONC_LOG_FORMAT",
    LOG_PREFIX + "[%(asctime)s][PID:%(process)d][%(levelname)s][%(name)s] %(message)s",
)


REST_API_PORT: int = -1


PENDING_TASKS_SOFT_LIMIT: int = -1
SHORT_JOB_MAX_TIME:float = 0


ONLY_CHOOSEN_REST_API_VERSION: bool = False
REST_API_VERSION: int = -1


MAX_THREADS_PER_TASK: int = 1
MAX_WORKER_THREADS: int = 1 


DEFAULT_TO_HUMAN_READABLE_NUMBERS = True


ENABLE_AUTH = True

USER_DEFINED_TOKEN = None

PLUGIN_PATHS: List[str] = []

DB_PATH = "JuMonC.db"

CACHE_DEFAULT_ENTRIES_PER_PAGE = 10

DATETIME_FORMAT = "%d.%m.%Y, %H:%M:%S"

SSL_ENABLED = False
SSL_MODE:Optional[Union[str,Tuple[str,str]]] = None



import logging

from typing import Dict, List
from types import ModuleType

import importlib.util

from JuMonC import settings


logger = logging.getLogger(__name__)


Plugin_modules: List[ModuleType] = []
Plugin_REST_paths: Dict[str, ModuleType] = {}


def startPlugins() -> None:
    for pluginPath in settings.PLUGIN_PATHS:
        try:
            spec = importlib.util.spec_from_file_location("module.name", pluginPath)
            if spec is not None:
                module = importlib.util.module_from_spec(spec)
                loader = spec.loader
                if loader is not None:
                    loader.exec_module(module)
            
                Plugin_modules.append(module)
        except Exception:
            logging.warning("User plugin \"%s\" can not be imported", pluginPath)
            

def gatherRESTpaths() -> None:
    for plugin in Plugin_modules:
        try:
            RESTpaths: List[str] = plugin.getNeededRESTpaths()
            for path in RESTpaths:
                Plugin_REST_paths[path] = plugin
        except Exception:
            logging.warning("User plugin \"%s\" does not supply a (correct) getNeededRESTpaths() function ", str(plugin))
            

def _removePathsAlreadyInUse() -> None:
    #pylint: disable=import-outside-toplevel
    from JuMonC.handlers.base import RESTAPI
    #pylint: enable=import-outside-toplevel
    
    routes: List[str] = [] # routes already registered
    removeKeys: List[str] = []
    
    for rule in RESTAPI.url_map.iter_rules():
        routes.append(rule.rule)
        
    for (key, module) in Plugin_REST_paths.items():
        if key in routes:
            logging.warning("User plugin has a path conflict, path: \"%s\" plugin \"%s\"", str(key), str(module))
            removeKeys.append(key)
            
    for key in removeKeys:
        del Plugin_REST_paths[key]
   

def registerRESTpaths() -> None:
    _removePathsAlreadyInUse()
    
    for (path, module) in Plugin_REST_paths.items():
        module.registerPath(path = path)



import logging

from flask import abort, jsonify, make_response, Response, request

from typing import Dict, List, Union


from JuMonC.handlers.base import api_version_path, check_version, RESTAPI
from JuMonC.models import pluginInformation
from JuMonC.authentication import scopes
from JuMonC.authentication.check import check_auth
from JuMonC.tasks import LinuxNetwork
from JuMonC import settings

import JuMonC.models.cache.helper as cache


logger = logging.getLogger(__name__)

links: List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []
status_links: List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []

network_path = "/network"

@RESTAPI.route(api_version_path + network_path, methods=["GET"])
@check_version
@check_auth(scopes["see_links"])
def returnNetworkLinks(version: int) -> Response:
    logging.debug("Accessed /v%i/network/", version)
    return make_response(jsonify(sorted(links, key=lambda dic: dic['link'])), 200)

@RESTAPI.route(api_version_path + network_path + "/status", methods=["GET"])
@check_version
@check_auth(scopes["see_links"])
def returnNetworkStatusLinks(version: int) -> Response:
    logging.debug("Accessed /v%i/network/status", version)
    return make_response(jsonify(sorted(status_links, key=lambda dic: dic['link'])), 200)

def registerRestApiPaths(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    links.append(registerStatusLinks(version))
    return {
        "link": "/v" + str(version) + network_path,
        "isOptional": False,
        "description": "Gather information concerning the network",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }

def registerStatusLinks(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    if pluginInformation.network_is_working:
        types = LinuxNetwork.plugin.getAvaiableDataTypes()
        descriptions = LinuxNetwork.plugin.getAvaiableDataTypesDescriptions()
        types.append("All")
        descriptions.append("get all network status information")
        for i, typeStr in enumerate(types):
            status_links.append({
            "link": "/v" + str(version) + network_path + "/status/" + typeStr,
            "isOptional": True,
            "description": descriptions[i],
            "parameters": [
                {"name": "duration",
                "description": "When this parameter is not present the raw data value(s) will be provided, otherwise the average value per second"},
                {"name": "interface",
                "description": ("When this parameter is not present information for all network interfaces are collected, otherwise only for the specified " + 
                     "interface. Avaiable interfaces: " + str(LinuxNetwork.plugin.getAvaiableInterfaces()))},
                {"name": "humanReadable",
                "description": "For 'True' convert to better readable numbers, for 'False' return actual number. " + 
                 "If not set to valid value, uses default value(" + str(settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS) + ")."},
                {"name": "node",
                "description": "When this parameter is not present the average data of all nodes is presented, otherwise only from the choosen node"},
                {"name": "token",
                "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
            })
        
        @RESTAPI.route(api_version_path + network_path + "/status/<string:datatype>", methods=["GET"])
        @check_version
        @check_auth(scopes["compute_data"])
        def returnLinuxNetworkStatus(version: int, datatype:str) -> Response:
            interface = request.args.get('interface', default = "all", type = str)
            duration = request.args.get('duration', default = -1.0, type = float)
            humanReadable = request.args.get('humanReadable', default = settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS, type = settings.helpers.parse_boolean)
            logging.debug("Accessed /v%i/network/status/%s", version, datatype)
            datatype = datatype.lower().capitalize()
            dataTypes = LinuxNetwork.plugin.getAvaiableDataTypes()
            dataTypes.append("All")
            if datatype in dataTypes:
                cache_id = cache.add_cache_entry("/v" + str(version) + "/network/status/" + datatype)
                if duration > 0:
                    cache.addParameter(cache_id, "duration", str(duration))
                cache.addParameter(cache_id, "interface", interface)
                data = LinuxNetwork.plugin.getData(dataType = datatype,
                                                        duration = duration,
                                                        interface = interface,
                                                        overrideHumanReadableWithValue = humanReadable)
                for data_element in data:
                    interface_str:Union[str, Dict[str, Union[int, float, str]]] = data_element["interface"]
                    if not isinstance(interface_str, str):
                        logging.warning("Unexpected datatype in passing networkdata into cache")
                        continue
                    for result_part_name, part in data_element.items():
                        if result_part_name == "interface":
                            continue
                        if isinstance(part, dict):
                            for result_name, result in part.items():
                                cache.addResult(cache_id, interface_str + "_" + result_part_name + "_" + result_name, result)
                
                return make_response(jsonify(data, 200))
            abort(404)
            return make_response("",404)
        
    return {
        "link": "/v" + str(version) + network_path + "/status",
        "isOptional": False,
        "description": "Gather information concerning the network status",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}
        ]
    }



import logging

from flask import Flask, abort, make_response, Response
import flask_login  # type: ignore

from functools import wraps

from typing import Callable, Any


from JuMonC import settings
from JuMonC.helpers.generateToken import generateToken


logger = logging.getLogger(__name__)

RESTAPI = Flask("JuMonC.handlers")
RESTAPI.config['JSON_SORT_KEYS'] = False
RESTAPI.url_map.strict_slashes = False



RESTAPI.config['SECRET_KEY'] = generateToken()
login_manager = flask_login.LoginManager()
login_manager.init_app(RESTAPI)


if settings.ONLY_CHOOSEN_REST_API_VERSION:
    start_version = settings.REST_API_VERSION
else:
    start_version = 1
end_version = settings.REST_API_VERSION

api_version_path = "/v<int:version>"


def check_version(func: Callable[..., Response] ) -> Callable[..., Response]:
    @wraps(func)
    def decorated_function(*args: Any, **kwargs: Any) -> Response:
        if kwargs["version"] >= start_version and kwargs["version"] <= end_version:
            return func(*args, **kwargs)
        abort(404)
        return make_response("",404)
            
    return decorated_function


import logging

import mpi4py
from mpi4py import MPI

from threading import Thread

import sys

from typing import Any


from JuMonC import settings
from JuMonC._version import __version__
from JuMonC.helpers.startup import communicateAvaiablePlugins
from JuMonC.helpers.cmdArguments import parseCMDOptions
from JuMonC.models.plugins import gatherRESTpaths, startPlugins, registerRESTpaths



logger = logging.getLogger(__name__)
logging.info("Running JuMonC with version: %s", __version__)

mpi4py.rc.threads = True
mpi4py.rc.thread_level = "funneled"

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def startJuMonC() -> None:
    parseCMDOptions(sys.argv[1:])
    startPlugins()
    communicateAvaiablePlugins()

    #pylint: disable=import-outside-toplevel
    if rank == 0:
        from JuMonC.handlers.base import RESTAPI
        from JuMonC.handlers import main
    
        from JuMonC.tasks import mpiroot
        from JuMonC.models.cache.database import Base, engine, db_session
        
        gatherRESTpaths()
    
        main.registerRestApiPaths()
        registerRESTpaths()

        if settings.SSL_ENABLED:
            JuMonC_SSL = settings.SSL_MODE
        else:
            JuMonC_SSL = None

        flask_thread = Thread(target=RESTAPI.run, kwargs={'debug': True, 'port': settings.REST_API_PORT, 'use_reloader': False, 'ssl_context': JuMonC_SSL})
        flask_thread.start()
        
        from JuMonC.models.cache import dbmodel
        Base.metadata.create_all(bind=engine)
        dbmodel.check_db_version()

        @RESTAPI.teardown_appcontext
        def shutdown_session(exception:Any = None) -> None:
            db_session.remove()
            if exception:
                logging.warning("DB connection close caused exception: %s", str(exception))
        
    
    
        mpiroot.waitForCommands()
    
    
        flask_thread.join()
    
    else:
        from JuMonC.tasks import mpihandler
    
        mpihandler.waitForCommands()
    #pylint: enable=import-outside-toplevel
        



import logging

import argparse

import sys

from typing import List


from JuMonC import settings
from JuMonC import reset_logging


logger = logging.getLogger(__name__)



def parseCMDOptions(args: List[str] = None) -> None:
    if args is None:
        args = []

    parser = setupParser()
    evaluateArgs(parser, args)



def setupParser() -> argparse.ArgumentParser:
    #pylint: disable=import-outside-toplevel
    from JuMonC._version import __DB_version__, __REST_version__, REST_version_info, __version__
    #pylint: enable=import-outside-toplevel

    parser = argparse.ArgumentParser(description=("JuMonC is the Jülich Monitoring and Control programm, "
                                                  "it allows to monitore your running simulations and access avaiable data using a REST-API"),
                                     prog="app.py",)

    parser.add_argument("--CACHE_DEFAULT_ENTRIES_PER_PAGE" ,
                       dest="CACHE_DEFAULT_ENTRIES_PER_PAGE",
                       help="Number of cache entries that are on one page of /cache/list by default",
                       default=10,
                       type=int)
    parser.add_argument("--DATETIME_FORMAT" ,
                       dest="DATETIME_FORMAT",
                       help=("Datetime format that will be used in the REST-API, "
                           "see: https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior"),
                       default="%d.%m.%Y, %H:%M:%S",
                       type=str)
    parser.add_argument("--DB_PATH" ,
                       dest="DB_PATH",
                       help=("Path and filename to either an existing JuMonC database, that wil lbe used, or a new one that will be created"),
                       default="JuMonC.db",
                       type=str)
    parser.add_argument("--DONT_DEFAULT_TO_HUMAN_READABLE_NUMBERS", 
                       dest="DONT_DEFAULT_TO_HUMAN_READABLE_NUMBERS", 
                       help="Sets wether numbers are converted into smaller numbers by default, can be overwritten for each API call", 
                       default=True, 
                       action='store_false')
    parser.add_argument("--LOG_FORMAT", 
                       dest="LOG_FORMAT", 
                       help="Set log format, usable values are the values supported by logging", 
                       default="[%(asctime)s][PID:%(process)d][%(levelname)s][%(name)s] %(message)s", type=ascii)
    parser.add_argument("--LOG_LEVEL", 
                       dest="LOG_LEVEL", 
                       help="Set the log level used by the logger", 
                       default="INFO", 
                       type=ascii, 
                       choices=["'ERROR'", "'WARN'", "'INFO'", "'DEBUG'"])
    parser.add_argument("--LOG_STDOUT", 
                       dest="LOG_STDOUT", 
                       help="If used log to stdout, otherwise to stderr", 
                       default=False, 
                       action='store_true')
    parser.add_argument("--LOG_PREFIX", 
                       dest="LOG_PREFIX", 
                       help="Set a prefix that will be prefaced to every logging output", 
                       default="", 
                       type=ascii)
    parser.add_argument("--MAX_WORKER_THREADS", 
                       dest="MAX_WORKER_THREADS", 
                       help="Limits the number of worker threads that work on the actual tasks at once", 
                       default=4, 
                       type=int)
    parser.add_argument("--ONLY_CHOOSEN_REST_API_VERSION", 
                       dest="ONLY_CHOOSEN_REST_API_VERSION", 
                       help="If set will only provide one version of the api links", 
                       default=False, 
                       action='store_true')
    parser.add_argument("--PENDING_TASKS_SOFT_LIMIT", 
                       dest="PENDING_TASKS_SOFT_LIMIT", 
                       help="Limits tasks being added by the REST-API, to not have more than PENDING_TASKS_SOFT_LIMIT tasks waiting", 
                       default=100, 
                       type=int)
    parser.add_argument("--PLUGIN_PATHS", 
                       dest="PLUGIN_PATHS", 
                       help="Paths to JuMonC plugins, multiple values allowed", 
                       default=[],
                       nargs='*',
                       type=str)
    parser.add_argument("-p", "--REST_API_PORT", 
                       dest="REST_API_PORT", 
                       help="Choose a port that the REST-API will be listening on", 
                       default=12121, 
                       type=int, 
                       metavar="[1024-65535]")
    parser.add_argument("--REST_API_VERSION", 
                       dest="REST_API_VERSION", 
                       help=("Choose a major version of the rest api. Depending on ONLY_CHOOSEN_REST_API_VERSION, "
                       "only this version, or all versions up to this version will be avaiable"), 
                       default=REST_version_info[0], 
                       choices=range(REST_version_info[0]+1), 
                       type=int)
    parser.add_argument("--SHORT_JOB_MAX_TIME", 
                       dest="SHORT_JOB_MAX_TIME", 
                       help=("Short jobs will be executed rigth away and return results directly via REST-API, "
                       "blocking all other mpi communication in between [s]"), 
                       default=0.1, 
                       type=float)
    parser.add_argument("--SSL_ENABLED",
                       dest="SSL_ENABLED",
                       help=("You are able to use SSL encrypted connections, by enabeling with this flag you are using adhoc"
                       "certificates. For further information see https://gitlab.jsc.fz-juelich.de/coec/jumonc#encryption"),
                       default=False,
                       action='store_true')
    parser.add_argument("--SSL_CERT",
                       dest="SSL_CERT",
                       help=("Supply a certificate to use for the SSL connection, can only be used with --SSL_ENABLED and --SSL_KEY."
                       "For further information see https://gitlab.jsc.fz-juelich.de/coec/jumonc#encryption"),
                       default=None,
                       type=str)
    parser.add_argument("--SSL_KEY",
                       dest="SSL_KEY",
                       help=("Supply a key to use for the SSL connection, can only be used with --SSL_ENABLED and --SSL_CERT."
                       "For further information see https://gitlab.jsc.fz-juelich.de/coec/jumonc#encryption"),
                       default=None,
                       type=str)
    parser.add_argument("--USER_DEFINED_TOKEN", 
                       dest="USER_DEFINED_TOKEN", 
                       help=("Define one additional token with scope level, separate multiple tokens by ;"
                       "Example \"--USER_DEFINED_TOKEN=12345678:100\""), 
                       default=None, 
                       type=str)
    parser.add_argument("-v", 
                       "--version", 
                       help="Print Version number of JuMonC", 
                       action='version', 
                       version=f'JuMonC\'s {__version__},\n REST-API\'s {__REST_version__},\n DB\'s {__DB_version__}')
    
    return parser





def evaluateArgs(parser: argparse.ArgumentParser, args: List[str]) -> None:
    parsed = parser.parse_args(args)

    evaluateLogArgs(parsed)
    evaluateRESTAPIArgs(parsed)
    evaluateDBArgs(parsed)
    evaluateThreadingArgs(parsed)
    evaluateSecurityArgs(parsed)
    evaluateMiscellaneousArgs(parsed)



def evaluateLogArgs(parsed:argparse.Namespace) -> None:
          
    #set new logging options first, to then use them
    LOG_LEVEL = parsed.LOG_LEVEL[1:-1]
    if settings.LOG_LEVEL != LOG_LEVEL:
        settings.LOG_LEVEL = LOG_LEVEL
        logging.warning("Changing LOG_LEVEL to %s", settings.LOG_LEVEL)
        reset_logging()
    
    if settings.LOG_STDOUT != parsed.LOG_STDOUT:
        settings.LOG_STDOUT = parsed.LOG_STDOUT
        logging.warning("Changing LOG_STDOUT to %s", str(settings.LOG_STDOUT))
        reset_logging()
    
    LOG_PREFIX = parsed.LOG_PREFIX[1:-1]
    if settings.LOG_PREFIX != LOG_PREFIX:
        settings.LOG_PREFIX = LOG_PREFIX
        logging.warning("Changing LOG_PREFIX to %s", settings.LOG_PREFIX)
        reset_logging()
    
    LOG_FORMAT = parsed.LOG_FORMAT[1:-1]
    if settings.LOG_FORMAT != LOG_FORMAT:
        settings.LOG_FORMAT = LOG_FORMAT
        logging.warning("Changing LOG_FORMAT to %s", settings.LOG_FORMAT)
        reset_logging()


def evaluateRESTAPIArgs(parsed:argparse.Namespace) -> None:
    #pylint: disable=import-outside-toplevel
    from JuMonC._version import REST_version_info
    #pylint: enable=import-outside-toplevel
    
    settings.ONLY_CHOOSEN_REST_API_VERSION = parsed.ONLY_CHOOSEN_REST_API_VERSION
    logging.info("Set ONLY_CHOOSEN_REST_API_VERSION to %s", str(settings.ONLY_CHOOSEN_REST_API_VERSION))

    if parsed.REST_API_PORT >=1024 and parsed.REST_API_PORT<=65535:
        settings.REST_API_PORT = parsed.REST_API_PORT
        logging.info("Set REST_API_PORT to %s", str(settings.REST_API_PORT))
    else:
        logging.error("Invalid value for REST_API_PORT: %s%s", str(settings.REST_API_PORT), ", needs to be between 1024 and 65535")
        sys.exit(-1)

    if parsed.REST_API_VERSION >= 1 and parsed.REST_API_VERSION <= REST_version_info[0]:
        settings.REST_API_VERSION = parsed.REST_API_VERSION
        logging.info("Set REST_API_VERSION to %s", str(settings.REST_API_VERSION))
    else:
        settings.REST_API_VERSION = REST_version_info[0]
        logging.warning("Invalid value for REST_API_VERSION: %s%s%s%s",
                        str(parsed.REST_API_VERSION), 
                        ", needs to be at least 1! Set to ",
                        str(settings.REST_API_VERSION ),
                        " now.")

    settings.DATETIME_FORMAT = parsed.DATETIME_FORMAT
    logging.info("Set DATETIME_FORMAT to %s", str(settings.DATETIME_FORMAT ))

    settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS = parsed.DONT_DEFAULT_TO_HUMAN_READABLE_NUMBERS
    logging.info("Set DEFAULT_TO_HUMAN_READABLE_NUMBERS to %s", str(settings.DEFAULT_TO_HUMAN_READABLE_NUMBERS))

    settings.CACHE_DEFAULT_ENTRIES_PER_PAGE  = parsed.CACHE_DEFAULT_ENTRIES_PER_PAGE 
    logging.info("Set CACHE_DEFAULT_ENTRIES_PER_PAGE  to %s", str(settings.CACHE_DEFAULT_ENTRIES_PER_PAGE ))


def evaluateDBArgs(parsed:argparse.Namespace) -> None:

    settings.DB_PATH = parsed.DB_PATH
    logging.info("Set DB_PATH to %s", str(settings.DB_PATH ))



def evaluateThreadingArgs(parsed:argparse.Namespace) -> None:

    if parsed.MAX_WORKER_THREADS >= 1:
        settings.MAX_WORKER_THREADS = parsed.MAX_WORKER_THREADS
        logging.info("Set MAX_WORKER_THREADS to %s", str(settings.MAX_WORKER_THREADS))
    else:
        settings.MAX_WORKER_THREADS = 1
        logging.warning("Invalid value for MAX_WORKER_THREADS: %s%s%s%s",
                        str(parsed.MAX_WORKER_THREADS), 
                        ", needs to be at valid version! Set to ",
                        str(settings.MAX_WORKER_THREADS ),
                        " now.")

    settings.SHORT_JOB_MAX_TIME = parsed.SHORT_JOB_MAX_TIME
    logging.info("Set SHORT_JOB_MAX_TIME to %s", str(settings.SHORT_JOB_MAX_TIME))

    if parsed.PENDING_TASKS_SOFT_LIMIT >= 1:
        settings.PENDING_TASKS_SOFT_LIMIT = parsed.PENDING_TASKS_SOFT_LIMIT
        logging.info("Set PENDING_TASKS_SOFT_LIMIT to %s", str(settings.PENDING_TASKS_SOFT_LIMIT))
    else:
        settings.PENDING_TASKS_SOFT_LIMIT = 1
        logging.warning("Invalid value for PENDING_TASKS_SOFT_LIMIT: %s%s%s%s",
                        str(parsed.PENDING_TASKS_SOFT_LIMIT), 
                        ", needs to be at valid version! Set to ",
                        str(settings.PENDING_TASKS_SOFT_LIMIT ),
                        " now.")



def evaluateSecurityArgs(parsed:argparse.Namespace) -> None:

    # SSL configuration, due to security concerns, JuMonC wil lbe stopped in case of incomplet configuration
    if parsed.SSL_ENABLED:
        if parsed.SSL_CERT or parsed.SSL_KEY:
            if parsed.SSL_CERT and parsed.SSL_KEY:
                settings.SSL_MODE = (parsed.SSL_CERT, parsed.SSL_KEY)
                logging.warning("Enabling SSL connection with user certificate and key")
            else:
                logging.warning("Cert: %s, Key: %s", str(parsed.SSL_CERT), str(parsed.SSL_KEY))
                logging.error(("For use of self supplied certificates JuMonC needs both the certificate and key.",
                                  " See: https://gitlab.jsc.fz-juelich.de/coec/jumonc#encryption"))
                sys.exit(-5)
        else:
            settings.SSL_MODE='adhoc'
            logging.warning("Enabling adhoc SSL connection")
        settings.SSL_ENABLED = parsed.SSL_ENABLED
    if (parsed.SSL_CERT or parsed.SSL_KEY) and not parsed.SSL_ENABLED:
        logging.error(("For use of self supplied certificates JuMonC needs the certificate, the key and SSl needs to be enabled.",
                          " See: https://gitlab.jsc.fz-juelich.de/coec/jumonc#encryption"))
        sys.exit(-5)
    
    settings.USER_DEFINED_TOKEN = parsed.USER_DEFINED_TOKEN
    logging.info("Set USER_DEFINED_TOKEN to %s", str(settings.USER_DEFINED_TOKEN))



def evaluateMiscellaneousArgs(parsed:argparse.Namespace) -> None:

    settings.PLUGIN_PATHS.extend(parsed.PLUGIN_PATHS)
    logging.info("Set PLUGIN_PATHS to %s", str(settings.PLUGIN_PATHS))


import logging

from flask import jsonify, make_response, Response, request

from typing import Dict, List, Union

from sqlalchemy import desc


from JuMonC.handlers.base import api_version_path, check_version, RESTAPI
from JuMonC.authentication import scopes
from JuMonC.authentication.check import check_auth
from JuMonC import settings

import JuMonC.models.cache.dbmodel as cache_model

from JuMonC.models import pluginInformation


logger = logging.getLogger(__name__)

links:List[Dict[str, Union[bool, str, List[Dict[str, str]]]]] = []

cache_path = "/cache"

@RESTAPI.route(api_version_path + cache_path, methods=["GET"])
@check_version
@check_auth(scopes["see_links"])
def returnCacheLinks(version: int) -> Response:
    logging.debug("Accessed /v%i/cache/", version)
    return make_response(jsonify(sorted(links, key=lambda dic: dic['link'])), 200)

def registerRestApiPaths(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    if pluginInformation.flask_path_reg is False:
        links.append(registerCacheList(version))
        links.append(registerCacheID(version))
    pluginInformation.flask_path_reg = True
    return {
        "link": "/v" + str(version) + cache_path,
        "isOptional": False,
        "description": "Retrieve old results",
        "parameters": [
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }


def registerCacheList(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
        
    @RESTAPI.route(api_version_path + cache_path + "/list", methods=["GET"])
    @check_version
    @check_auth(scopes["retrieve_simulation_data"])
    def returnCacheList(version: int) -> Response:
        entries = request.args.get('entries', default = settings.CACHE_DEFAULT_ENTRIES_PER_PAGE, type = int)
        page = request.args.get('page', default = 1, type = int)
        logging.debug("Accessed /v%i/cache/list/", version)

        if page < 1:
            return make_response("Pages start with 1, you set page number to: " + str(page), 422)

        
        start = (page-1)*entries
        cache_data = cache_model.CacheEntry.query.order_by(desc(cache_model.CacheEntry.cache_id)).offset(start).limit(entries)
        
        data:List[Dict[str, Union[int, float,str]]] = []
        index:Dict[str, Union[int, float,str]] = {"page": page,
                     "description": "Showing cache elements " + str(start) + " to " + str(start+entries) + ", most recent results first",
                    }
        if entries == settings.CACHE_DEFAULT_ENTRIES_PER_PAGE:
            if page > 1:
                index["previous page"] = "/v" + str(version) + "cache/list/?page=" + str(page-1)
            index["next page"] = "/v" + str(version) + "/cache/list/?page=" + str(page+1)
        else:
            if page > 1:
                index["previous page"] = "/v" + str(version) + "cache/list/?entries=" + str(entries) + "&page=" + str(page-1)
            index["next page"] = "/v" + str(version) + "/cache/list/?entries=" + str(entries) + "&page=" + str(page+1)
        data.append(index)
            
        for element in cache_data:
            data.append({"ID": element.cache_id,
                         "time": element.time.strftime(settings.DATETIME_FORMAT),
                         "link": "/v" + str(version) + "/cache/id/?id=" + str(element.cache_id),
                         "API_path": element.API_path})
        
        return make_response(jsonify(data), 200)
        
    return {
        "link": "/v" + str(version) + cache_path + "/list",
        "isOptional": False,
        "description": "List old results that can be retrieved, if to many results are avaiable, it will be split into pages",
        "parameters": [
            {"name": "page",
             "description": "Which page will be used"},
            {"name": "entries",
             "description": "How many entries are per page"},
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }
        


def registerCacheID(version: int) -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    
    @RESTAPI.route(api_version_path + cache_path + "/id", methods=["GET"])
    @check_version
    @check_auth(scopes["retrieve_simulation_data"])
    def returnCacheID(version: int) -> Response:
        c_id = request.args.get('id', type = int)
        logging.debug("Accessed /v%i/cache/id/", version)
        
        entry = cache_model.CacheEntry.query.filter_by(cache_id=c_id).first()
        if not entry:
            return make_response(("Cache entry with ID ",
                                  str(c_id),
                                  "could not be found. You can use /v",
                                  str(version),
                                  "/cache/list to see all avaiable entries"),
                            422)
            
        data:Dict[str, Union[bool, str, Dict[str, str]]] = {}
        data["Used API path"] = entry.API_path
        data["time"] = str(entry.time.strftime(settings.DATETIME_FORMAT))
            
        parameters:Dict[str, str] = {}
        for parameter in entry.parameters:
            parameters[parameter.parameter_name] = str(parameter.parameter_value)
        data["parameter"] = parameters
            
        results:Dict[str, str] = {}
        for result in entry.results_int:
            results[result.result_name] = str(result.result)
        for result in entry.results_float:
            results[result.result_name] = str(result.result)
        for result in entry.results_str:
            results[result.result_name] = result.result
        data["results"] = results  
            
        return make_response(jsonify(data), 200)
        
    return {
        "link": "/v" + str(version) + cache_path + "/id",
        "isOptional": False,
        "description": "Retrieve complete cache entry by ID",
        "parameters": [
            {"name": "id",
             "description": "ID of the cahce entry that shall be retrieved"},
            {"name": "token",
             "description": "Supply a token that shows you are allowed to access this link (or login once using /login)"}]
    }
        



# Module version stage suffix map
_specifier_ = {'alpha': 'a', 'beta': 'b', 'candidate': 'rc', 'final': ''}


# Module version
version_info = (0, 6, 0, 'candidate', 2)

version_end = '' if version_info[3] == 'final' else _specifier_[version_info[3]]+str(version_info[4])
__version__ = f'{version_info[0]}.{version_info[1]}.{version_info[2]}{version_end}' 


# REST-API version
REST_version_info = (1, 0, 1, 'beta', 0)

REST_version_end = '' if REST_version_info[3] == 'final' else _specifier_[REST_version_info[3]]+str(REST_version_info[4])
__REST_version__ = f'{REST_version_info[0]}.{REST_version_info[1]}.{REST_version_info[2]}{REST_version_end}' 


# DB version
DB_version_info = (1, 0, 0, 'final', 0)

DB_version_end = '' if DB_version_info[3] == 'final' else _specifier_[DB_version_info[3]]+str(DB_version_info[4])
__DB_version__ = f'{DB_version_info[0]}.{DB_version_info[1]}.{DB_version_info[2]}{DB_version_end}' 



from typing import Dict, List, Union

from flask import jsonify, make_response, request, Response
from flask_login import login_user, login_required, current_user, logout_user  # type: ignore


from JuMonC.handlers.base import RESTAPI
from JuMonC.authentication import tokens, scopes, scope_name
from JuMonC.authentication.check import check_auth


@RESTAPI.route("/login", methods=["GET"])
def login() -> Response:
    token = request.args.get('token', default = None, type = str)
    scope = tokens.getTokenScope(token)
    if scope is None:
        return make_response(jsonify("The supplied token is not valid"), 400)
    
    login_user(scope)
    response_body = "Logged in with scope: " + scope_name[tokens.tokens[scope.id]]
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/scope", methods=["GET"])
@login_required
def scope() -> Response:
    response_body = "Scope: " + scope_name[tokens.tokens[current_user.id]]
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/logout", methods=["GET"])
@login_required
def logout() -> Response:
    response_body = "Logged out of scope: " + scope_name[tokens.tokens[current_user.id]]
    logout_user()
    return make_response(jsonify(response_body), 200)


def registerLogin() -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    return {
        "link": "/login",
        "isOptional": False,
        "description": "Calling this link allows you to login using the appropiate scope_token",
        "parameters": [
            {"name": "token",
             "description": "A token defining the access scope"}
        ]
    }
    
def registerLogout() -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    return {
        "link": "/logout",
        "isOptional": False,
        "description": "Calling this link removes the use of the scope token provided in the login function",
        "parameters": [
            {}
        ]
    }
    
def registerScope() -> Dict[str, Union[bool, str, List[Dict[str, str]]]]:
    return {
        "link": "/scope",
        "isOptional": False,
        "description": "Calling this link shows the scope of this login",
        "parameters": [
            {}
        ]
    }


@RESTAPI.route("/test_auth_see_links", methods=["GET"])
@check_auth(scopes["see_links"])
def test_auth_see_links() -> Response:
    response_body = "Your scope: " + scope_name[tokens.tokens[current_user.id]] + " gives authorisation to visit the see_links scope"
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/test_auth_retrieve_data", methods=["GET"])
@check_auth(scopes["retrieve_data"])
def test_auth_retrieve_data() -> Response:
    response_body = "Your scope: " + scope_name[tokens.tokens[current_user.id]] + " gives authorisation to visit the retrieve_data scope"
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/test_auth_compute_data", methods=["GET"])
@check_auth(scopes["compute_data"])
def test_auth_compute_data() -> Response:
    response_body = "Your scope: " + scope_name[tokens.tokens[current_user.id]] + " gives authorisation to visit the compute_data scope"
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/test_auth_retrieve_simulation_data", methods=["GET"])
@check_auth(scopes["retrieve_simulation_data"])
def test_auth_retrieve_simulation_data() -> Response:
    response_body = "Your scope: " + scope_name[tokens.tokens[current_user.id]] + " gives authorisation to visit the retrieve_simulation_data scope"
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/test_auth_compute_simulation_data", methods=["GET"])
@check_auth(scopes["compute_simulation_data"])
def test_auth_compute_simulation_data() -> Response:
    response_body = "Your scope: " + scope_name[tokens.tokens[current_user.id]] + " gives authorisation to visit the compute_simulation_data scope"
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/test_auth_full", methods=["GET"])
@check_auth(scopes["full"])
def test_auth_full() -> Response:
    response_body = "Your scope: " + scope_name[tokens.tokens[current_user.id]] + " gives authorisation to visit the full scope"
    return make_response(jsonify(response_body), 200)


# General commands 1-10000

finalize = 1


# Job commands 10001-20000



# CPU commands 20001-30000

meassure_cpu_load = 20001
retrieve_cpu_load = 20002

meassure_cpu_frequency = 20011


# GPU commands 30001-40000

meassure_gpu_load = 30001
retrieve_gpu_load = 30002

meassure_gpu_frequency = 30011


# Network commands 40001-50000



# IO commands 50001-60000



# General application commands 60001-70000



# Alya commands 70001-80000


from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from JuMonC import settings

engine = create_engine('sqlite:///' + settings.DB_PATH)
db_session = scoped_session(sessionmaker(autocommit=False,
                                         autoflush=False,
                                         bind=engine))
Base = declarative_base()
